program rtm2d
!****************************************************************************************************
!*** Main program for 2D Reverse-time migration.                                                  ***
!*** This program performs RTM using the rapid-expansion method with the pseudospectral approach  ***
!*** Author: Alejandro Cabrales. Stanford Exploration Project/Petroleos Mexicanos, 2015           ***
!****************************************************************************************************
use sep
use omp_lib
use weqsolver_mod 
use rtm_mod
use vshot_mod
use iso_c_binding
use base_trismt_mod
use lap_c_mod
implicit none
!include 'omp_lib.h'
!*************************************
!*** File handle and misc integers ***
!*************************************
integer :: nphase = 10
integer :: in_error, input_error, dummy, err
integer :: allocation_status  ! Return code from allocate statement
integer, dimension(8)  :: timer0, timer1
!*************************
!*** Indexes variables ***
!*************************
integer :: k, l, jt, jtCoarse, jtStep, jz, jx, jgx, jtraces, jgeof, jaux, jsources, jencod, jtsnap=1 
integer :: nt, ntData, ntSnap=0, ntstep, nts, nts_data, nx, nz, ngx, ngx_p, nxpad, ngeof, nkgx, nkz
integer :: ntraces, ntracesData, nsources, nbox, napLDta, napRDta, napLMdl, napRMdl, npts
integer :: nsx_enc, ngx_enc, nsuper_gath
integer :: src_snap, shot_pos, src_pad 
integer :: nthreads
integer :: first_cdp, last_cdp, start_cdp, end_cdp, first_source, last_source, first_geof, last_geof
integer :: start_jz, end_jz
integer :: pad, zpad = 30
integer :: thetac
integer :: scalco = 1
!********************************
!*** AASPI internal variables ***
!********************************
integer :: scalco_key, sx_key, gx_key, sy_key, gy_key, n_keys_data, n_keys_vel
integer :: gath_enc_key, super_key, encod_key
integer :: next_header_out, previous_trace, n_nonzero_samples
!**************************
!*** Allocatable arrays ***
!**************************
integer, dimension(:,:), allocatable :: header_in, source_index
integer, dimension(:),   allocatable :: jtraces_ind
integer, dimension(:),   allocatable :: shots_blended, super_shots_pos, super_gath, encod_vector
!******************************
!*** SEP external functions ***
!******************************
integer, external :: sseek_block, sreed
integer, external :: sep_get_val_headers, sep_get_header_format_tag, sep_get_number_keys, sep_get_key_index
!***********************************
!*** File handle and misc floats ***
!***********************************
real :: timer_sum=0.
!********************************************
!*** Sampling rates and Fourier variables ***
!********************************************
real :: dt, dtData, dtSnap, dx, dz, dsx, dgx, dgxOdx, dkgx, dkz
real :: tIntFact
real :: kgx_Nyq, kz_Nyq, kNyq, kz0, kzp , kcl, kcu
real :: ox_shift
real :: hsx_cur,hsx_next, hgx_cur, hgx_next
real :: scalcof
!**************************
!*** Sep real variables ***
!**************************
real :: d2, d3
real :: o1, o2, o3
real :: amp_rms, amp_mean
real :: min_amplitude = 99999.0  
real :: max_amplitude = -99999.0
!***********************************
!*** General migration variables ***
!***********************************
real :: t_end, z_start, z_end
real :: Vcte, Vmin, Vmax, eps_min, eps_max, delta_min, delta_max, theta_min, theta_max
real :: vel_fracc
real :: hmin
real :: fNyq, fc
real :: aperture, apertureL, apertureR
real :: jfx, cfl, denmax, denmax0, cos2thc, alpha
real :: tiny=0.01
!***********************************
!*** Useful auxiliary variables ***
!***********************************
real :: faux1, faux2, faux3 
!***********************
!*** Some parameters ***
!***********************
real, parameter :: pi = 3.1415926535897932 ! The Order of Chaos (Darren Aronofsky, 1998)
real, parameter :: UNDEFINED = -9999999.
real, parameter :: deg2rad = pi/180.
!**************************
!*** Allocatable arrays ***
!**************************
real, dimension(:),     allocatable :: x, sx, gx, kgx, sgx_max, kz, fkzsep
real, dimension(:,:),   allocatable :: sgx    ! Source coordinates
real, dimension(:,:),   allocatable :: vinput, V_int, V_smth, V_shot, Vshot0, V_shotr, V_int2
real, dimension(:,:),   allocatable :: fanis1, fanis2, fanis3, fanis4, fanis5
real, dimension(:,:),   allocatable :: epsShot, delShot
real, dimension(:,:),   allocatable :: lambP, lambM, dd2
real, dimension(:,:),   allocatable :: Vanis1, Vanis2
real, dimension(:,:),   allocatable :: V_shot_a1, V_shot_a2, V_shot_a3, V_shot_a4, V_shot_a5
real, dimension(:,:),   allocatable :: alp
real, dimension(:,:),   allocatable :: eps, delta, theta, sinth, costh, sin2th, cos2th
real, dimension(:),     allocatable :: rcv, s, src
real, dimension(:,:),   allocatable :: shot_gath  ! Input data trace by trace
real, dimension(:),     allocatable :: hoff
real, dimension(:,:),   allocatable :: I_corr, I_tomo, I_refl, I0, I90
real, dimension(:),     allocatable :: super_s  !
real, dimension(:,:),   allocatable :: pft, pftR, pftX, pftZ, pftPX, pftRX, pftPZ, pftRZ
real, dimension(:,:),   allocatable :: dpft, dpftP, dpftR
real, dimension(:,:),   allocatable :: prt, prtR, prtX, prtZ, prtPX, prtRX, prtPZ, prtRZ
real, dimension(:,:),   allocatable :: dprt, dprtP, dprtR
real, dimension(:,:),   allocatable :: pS, pftm1, pftm2
real, dimension(:,:),   allocatable :: dpftPm1, dpftPm2, dpftRm1, dpftRm2
real, dimension(:,:),   allocatable :: prtp1, prtp2
real, dimension(:,:),   allocatable :: dprtPp1, dprtPp2, dprtRp1, dprtRp2
real, dimension(:,:),   allocatable :: psnap, qft, qrt
real, dimension(:,:),   allocatable :: wft, wft1, wft2, wft1R, wft2R
real, dimension(:,:),   allocatable :: wrt, wrt1, wrt2, wrt1R, wrt2R
real, dimension(:,:),   allocatable :: a, b, denreal
real, dimension(:),     allocatable :: am
real, dimension(:,:),   allocatable :: num, num0, num90, den, den0, temp1, temp2, temp3
!************************************************************
!*** Logical variables to make some functions work or not ***
!************************************************************
logical :: anis, eps_f, del_f, th_f
logical :: shot_blend, snapsht, one_src, src_ascii
logical :: tti, energy, ispc, override_sx_header, jump_src
!**********************************
!*** Single character variables ***
!**********************************
character :: char_aux  ! Auxiliary character variable for input purposes
!*******************************
!*** SEP character variables ***
!*******************************
character(10) :: unit1, unit2, unit3
character(6) :: snapfile 
character(len=1024) :: wavelet_fn, tag_input 
character(:), allocatable :: snap_fn, snapr_fn 
!******************************
!*** Non-standard variables ***
!****************************** 
write(0,'(a)') ' '
write(0,'(a)') '****************************************************************************************'
write(0,'(a)') '*** 2D Reverse time migration program using Lax-Wendroff or pseudospectral methods     *'
write(0,'(a)') '*** Author: Alejandro Cabrales. Stanford Exploration Project/Petroleos Mexicanos, 2015 *'
write(0,'(a)') '****************************************************************************************'
write(0,'(a)') ' '
call sep_init()
call init_3d()
!---------------------------------------------------------------------------
call DATE_and_TIME(values=timer0)
!---------------------------------------------------------------------------
!*******************************
!* Read the input history file *
!*******************************
write(0,*) 'Read history and parameters'
!*** Parameters read from the command line ***
call from_param('snapfile',snapfile,'Project')
call from_param('vel_frac',vel_fracc,1.0)
call from_param('npts',npts,0)
call from_param('thetac',thetac,90)
call from_param('apertureL',apertureL,1000.0)
call from_param('apertureR',apertureR,1000.0)
call from_param('scalco_override',scalco,1)
call from_param('epsilon',char_aux,'n')
  eps_f = (char_aux == 'y')
call from_param('delta',char_aux,'n')
  del_f = (char_aux == 'y')
call from_param('theta',char_aux,'n')
  th_f = (char_aux == 'y')
call from_param('wavelet_fn',wavelet_fn)
call from_param('nthreads',nthreads,1)
call from_param('snapsht',char_aux,'n')
  snapsht = (char_aux == 'y')
call from_param('src_snap',src_snap,nsources)
call from_param('one_src',char_aux,'n')
  one_src = (char_aux == 'y')
call from_param('nts',nts_data,10)
call from_param('pad',pad,20)
call from_param('src_ascii',char_aux,'y')
  src_ascii = (char_aux == 'y')
call from_param('energy',char_aux,'n')
  energy = (char_aux == 'y')
call from_param('alpha',alpha,0.0)
call from_param('ispc',char_aux,'n')
  ispc = (char_aux == 'y')
call from_param('override_sx_header',char_aux,'n')
  override_sx_header = (char_aux == 'y')

cos2thc=cos(2.*thetac*deg2rad)

!if(snapsht .or. one_src) nthreads = 8

anis= .false.
if(eps_f .or. del_f) then
  anis = .true.
  write(0,*) 'read_history: Velocity model is anisotropic'
end if
!*** Parameters read from the history file ***
call from_history('n1',nz)
call from_history('n2',nx)
call from_history('o1',o1,0.00)
call from_history('o2',o2,0.00)
call from_history('d1',dz,0.00)
call from_history('d2',d2,0.00)
call from_history('d2',dx)
call from_history('unit1',unit1,'')
call from_history('unit2',unit2,'')
!*** Parameters read from the auxiliary files ***
call from_aux("datafile",'n1',ntData,0)
call from_aux("datafile",'n2',ntracesData,0)
call from_aux("datafile",'d1',dtData,0.00)
call from_aux("datafile",'nsx_enc',nsx_enc,1)
call from_aux("datafile",'ngx_enc',ngx_enc,1)
call from_aux("datafile",'nsuper_gath',nsuper_gath,1)
!*** ox_shift is used to shift all the coordinates to make them divisible
!*** by the sampling rate. 
call from_param('ox_shift',ox_shift,0.0)  
first_cdp = 1+floor((o2+ox_shift)/d2)
last_cdp = first_cdp+(nx-1) 
!*** More parameters read from the command line ***
t_end = (ntData-1)*dtData
z_start = o1
z_end = o1+dz*(nz-1)
start_jz = 1+floor(z_start/dz)
end_jz = 1+floor(z_end/dz)
write(0,*) 'Finishing reading history and parameters'
!********************************************
!*** Set the number of threads for OpenMP ***
!********************************************
call omp_SET_NUM_THREADS(nthreads)
!*****************************************
!* Finish reading the input history file *
!*****************************************
aperture = apertureL+apertureR
shot_blend = .false.
if(nsx_enc > 1) then
  write(0,*) 'Input data are supergathers'
  shot_blend = .true.
end if
!***********************
! Allocate main arrays *
!***********************
allocate(x(first_cdp:last_cdp),                     &
         V_int(start_jz:end_jz,first_cdp:last_cdp), &
         V_smth(start_jz:end_jz,first_cdp:last_cdp),&
         Vanis1(start_jz:end_jz,first_cdp:last_cdp),&
         Vanis2(start_jz:end_jz,first_cdp:last_cdp),&
         eps   (start_jz:end_jz,first_cdp:last_cdp),&
         delta (start_jz:end_jz,first_cdp:last_cdp),&
         theta (start_jz:end_jz,first_cdp:last_cdp),&
         sinth (start_jz:end_jz,first_cdp:last_cdp),&
         costh (start_jz:end_jz,first_cdp:last_cdp),&
         sin2th(start_jz:end_jz,first_cdp:last_cdp),&
         cos2th(start_jz:end_jz,first_cdp:last_cdp),&
         I_corr(start_jz:end_jz,first_cdp:last_cdp),&
         I_tomo(start_jz:end_jz,first_cdp:last_cdp),&
         I_refl(start_jz:end_jz,first_cdp:last_cdp),&
         I0    (start_jz:end_jz,first_cdp:last_cdp),&
         I90   (start_jz:end_jz,first_cdp:last_cdp),&
         temp3 (start_jz:end_jz,first_cdp:last_cdp),&
         STAT=allocation_status)
if(allocation_status /= 0) then
  write(0,*) 'Error! could not allocate main arrays!'
  call skull()
end if

if(shot_blend) then
  allocate(super_shots_pos(nsx_enc))
end if
if(shot_blend) then
  allocate(shots_blended(ntraces),&
     super_gath(ntraces),  &
     encod_vector(ntraces),  &
          STAT=allocation_status)
  if(allocation_status /= 0) then
    write(0,*) 'Error! could not allocate arrays for super gathers mode! ntraces = ',ntraces
    call skull()
    stop 666
  end if
end if
!**************************
!* Prepare image geometry *
!**************************
do jx = first_cdp,last_cdp
  x(jx) = jx*dx+ox_shift
end do
!**************************
!* Read the input headers *
!**************************
write(0,*) 'Reading the input headers'
dummy=sep_get_header_format_tag("datafile",tag_input)
if(dummy /= 0) then
  write(0,*) 'No dataset detected; aborting'
  call skull()
  stop 666
else
  if(sep_get_number_keys("datafile",n_keys_data) /= 0) then
    write(0,*) 'Error, cannot read in number of keys from the hff (.H@@) file!'
    input_error=input_error+1
  end if
  if(sep_get_key_index("datafile","sx",sx_key) /= 0) then
    write(0,*) 'error! sx key index does not exist on datafile file!'
    input_error=input_error+1
  end if
  if(sep_get_key_index("datafile","gx",gx_key) /= 0) then
    write(0,*) 'error! gx key index does not exist on datafile file!'
    input_error=input_error+1
  end if
  if(sep_get_key_index("datafile","sy",sy_key) /= 0) then
    write(0,*) 'error! sy key index does not exist on datafile file!'
    input_error=input_error+1
  end if
  if(sep_get_key_index("datafile","gy",gy_key) /= 0) then
    write(0,*) 'error! gy key index does not exist on datafile file!'
    input_error=input_error+1
  end if
  if(shot_blend) then
    if(sep_get_key_index("datafile","gath_enc",gath_enc_key) /= 0) then
      write(0,*) 'error! gath_enc key index does not exist on datafile file! Is it supershots?'
      input_error=input_error+1
    end if
    if(sep_get_key_index("datafile","super_gath",super_key) /= 0) then
      write(0,*) 'error! super key index does not exist on datafile file! Is it supershots?'
      input_error=input_error+1
    end if
    if(sep_get_key_index("datafile","encod_number",encod_key) /= 0) then
      write(0,*) 'error! encode key index does not exist on datafile file! Is it supershots?'
      input_error=input_error+1
    end if
  end if
  if(sep_get_key_index("datafile","scalco",scalco_key) /= 0) then
    write(0,*) 'error! scalco key index does not exist on datafile file!'
    input_error=input_error+1
  end if
  allocate(header_in(n_keys_data,ntracesData))
  header_in = 0
  if(sep_get_val_headers("datafile",1,ntracesData,header_in) /= 0) then
    write(0,*) 'could not read from input data headers!'
    call skull()
    stop 666
  end if
  !**************************
  !* Input geometry reading *
  !**************************
  if(scalco == 0) scalco = header_in(scalco_key,1)

  dsx = header_in(sx_key,2)-header_in(sx_key,1)
  dgx = header_in(gx_key,2)-header_in(gx_key,1)
  scalcof = scalco
  if(scalco < 0) then
    scalcof = -1./scalcof
  end if
  dgx = dgx*scalcof
  dgxOdx = ceiling(dgx/dx)
  if(dgxOdx == 1) then
    ntraces = ntracesData
  else
    ntraces = ntracesData*dgxOdx
  end if
  allocate(sx(ntraces),          &
           gx(ntraces),          &
           hoff(ntraces),        &
           jtraces_ind(ntraces))
  jump_src = .false.
  do jtraces = 1,ntraces
    jfx = 1+(jtraces-1)/dgxOdx
    jx = jfx
    jfx = jfx-jx
    if(jx >= 1 .and. jx < ntracesData) then
      hsx_cur  = header_in(sx_key,jx)*scalcof
      hsx_next = header_in(sx_key,jx+1)*scalcof
      hgx_cur  = header_in(gx_key,jx)*scalcof
      hgx_next = header_in(gx_key,jx+1)*scalcof
      sx(jtraces) = hsx_cur
      if(abs(hgx_next-hgx_cur) == dgx) then
        gx(jtraces) = ((1.0-jfx)*hgx_cur+jfx*hgx_next)
      elseif(jfx == 0.0) then
        gx(jtraces) = hgx_cur
        gx(jtraces+1) = hgx_cur+dx
      end if
    elseif(jx == ntracesData) then
      sx(jtraces)   = header_in(sx_key,jx)*scalcof
      gx(jtraces)   = header_in(gx_key,jx)*scalcof
    end if
  end do
  sx = sx+ox_shift
  gx = gx+ox_shift
  if(shot_blend) then
    shots_blended = header_in(gath_enc_key,:)
    super_gath = header_in(super_key,:)
    encod_vector = header_in(encod_key,:)
  end if
  deallocate(header_in)
end if
!***********************************************
!* Calculate offsets withing the data geometry *
!***********************************************
hoff = 0
do jtraces = 1,ntraces
  hoff(jtraces) = abs(gx(jtraces)-sx(jtraces))
end do
hmin = minval(hoff)
src_pad = floor(hmin/dx)
!***************************
!* Initialize FFTW threads *
!***************************
!if(nthreads /= 1) then
!  call sfftw_init_threads(err)
!  write(0,*) 'Initialize FFTW threads, ',err
!  if(err == 0) then
!    write(0,*) 'There is a problem initializing FFTW threads, err: ',err
!    call skull()
!    stop 666
!  end if 
!  call sfftw_plan_with_nthreads(INT(nthreads,c_int))
!end if
!*******************************
!* Some anisotropic parameters *
!*******************************
eps = 0
delta = 0
theta = 0
eps_max = 0
delta_max = 0
theta_max = 0
!*****************************
!* Reading the velocity file *
!*****************************
write(0,*) 'Reading velocities...'

allocate(vinput(start_jz:end_jz,first_cdp:last_cdp),STAT=allocation_status)
if(allocation_status /= 0) then
  write(0,*) 'Error! could not allocate velocity! start_jz, end_jz, first_cdp, last_cdp: ',&
                                                       start_jz, end_jz, first_cdp, last_cdp
  call skull()
  stop 666
end if
call sep_read(vinput)
do jx = first_cdp,last_cdp
  do jz = start_jz,end_jz
    if (vinput(jz,jx) <= 0) then
      write(0,*) 'Negative or zero input velocity value: ', vinput(jz,jx),jz,jx
      call skull()
      stop 666
      vinput(jz,jx) = vinput(jz,jx-1)
      write(0,*) 'New velocity value: ', vinput(jz,jx),jz,jx
    end if
  end do
  if (isnan(doT_PRODUCT(vinput(:,jx),vinput(:,jx)))) then
    write(0,*) 'NaN input velocity value '
    call skull()
    stop 666
  end if
end do

V_int = vel_fracc*vinput(:,first_cdp:last_cdp)
V_smth = V_int

deallocate(vinput)
write(0,*) 'Finish reading velocities...'

if(anis) then
  write(0,*) 'Reading anisotropy...'
  allocate(vinput(start_jz:end_jz,first_cdp:last_cdp),&
           STAT=allocation_status)
  if(allocation_status /= 0) then
    write(0,*) 'Error! could not allocate anisotropy! nz = ',nz
    call skull()
     stop 666
  end if
  vinput = 0
  if(eps_f) then
!*************************
!* Reading epsilon field *
!*************************
    call sep_read(vinput,"anis1")
    do jx = first_cdp,last_cdp
      if (minval(vinput(:,jx)) < 0) then
        write(0,*) 'Negative input epsilon value: ', minval(vinput(:,jx))
        call skull()
        stop 666
      end if
      if (isnan(doT_PRODUCT(vinput(:,jx),vinput(:,jx)))) then
        write(0,*) 'NaN input eps value '
        call skull()
        stop 666
      end if
    end do
    eps(:,first_cdp:last_cdp) = vel_fracc*vinput(:,first_cdp:last_cdp)
!    eps = eps+0.001*maxval(eps)
    write(0,*) 'Finish reading epsilon...'
  else
    eps = 0
  end if

  vinput = 0
  if(del_f) then
!***********************
!* Reading delta field *
!***********************
    call sep_read(vinput,"anis2")
    do jx = first_cdp,last_cdp
      if (isnan(dot_product(vinput(:,jx),vinput(:,jx)))) then
        write(0,*) 'NaN input delta value '
        call skull()
        stop 666
      end if
    end do
    delta(:,first_cdp:last_cdp) = vel_fracc*vinput(:,first_cdp:last_cdp)
    write(0,*) 'Finish reading delta...'
  else
    delta = 0
  end if

  if(th_f) then
    tti = .true.
!***********************
!* Reading theta field *
!***********************
    call sep_read(vinput,"anis3")
    do jx = first_cdp,last_cdp
      if (isnan(doT_PRODUCT(vinput(:,jx),vinput(:,jx)))) then
        write(0,*) 'NaN input theta value '
        call skull()
        stop 666
      end if
    end do
    theta(:,first_cdp:last_cdp) = vel_fracc*vinput(:,first_cdp:last_cdp)
    write(0,*) 'Finish reading theta...'
  else
    tti = .false.
    theta = 0
  end if
  deallocate(vinput)
  eps_min = minval(eps)
  eps_max = maxval(eps)
  delta_min = minval(delta)
  delta_max = maxval(delta)
  theta_min = minval(theta)
  theta_max = maxval(theta)
  if(theta_max /= 0. .OR. theta_min /= 0.) then
!    theta=abs(theta)
    do jgx = first_cdp,last_cdp
      do jz = 1,nz
        sinth(jz,jgx) = SIN(theta(jz,jgx)*deg2rad)
        sin2th(jz,jgx) = SIN(2.0*theta(jz,jgx)*deg2rad)
        costh(jz,jgx) = COS(theta(jz,jgx)*deg2rad)
        cos2th(jz,jgx) = COS(2.0*theta(jz,jgx)*deg2rad)
      end do
    end do
  else
    sinth = 0
    sin2th = 0
    costh = 1
    cos2th = 1
  end if
!*******************************************************************
!* Define anisotropic functions for decoupled anisotropy equations *
!*******************************************************************
  Vanis1 = eps
  Vanis2 = delta 
end if
deallocate(eps, delta) 
!********************************
!* Smoothing the velocity field *
!********************************
Vcte = SUM(V_smth(:,:))/(nx*nz)
if (Vcte <= 0) then
  write(0,*) ' Velocitiy is cero or negative, check input velocities! ',Vcte
  call skull()
  stop 666
end if
Vmin = minval(V_smth)
Vmax = maxval(V_smth)
!*******************************
! Preparing time domain arrays *
!*******************************
!*** Refine time sampling to meet Courant criterion
call from_param('tIntFact',tIntFact,1.0)
cfl = Vmax*dtData/min(dx,dz)
if(anis) then
  if(theta_min <= -45.) then
    faux3=sqrt(2.*(1.+2.*eps_max))
  else
    faux1=1+eps_max*(1.-sin(2.*theta_min*deg2rad))
    faux2=(eps_max-delta_max)*cos(2.*theta_min*deg2rad)**2
    faux3=sqrt(faux1+sqrt(faux1**2-faux2))
  end if
  k = ceiling(cfl*faux3)
!  tIntFact = cfl*faux3
!write(0,*) 'faux3: ',faux3
!  k = ceiling(cfl/0.65)
else
  k = ceiling(cfl/0.7)
end if
tIntFact = k
dt = dtData/tIntFact
jtsnap = 10
nt = 1+nint(t_end/dt)

snap_fn =  snapfile 
snapr_fn = snapfile //"r" 
allocate(src(ntData), s(nt),   &
         STAT=allocation_status)
if(allocation_status /= 0) then
  write(0,*) 'Error! could not allocate time dependent arrays!'
  call skull()
end if
src = 0
if(src_ascii) then
  nts_data = 0
  open(unit=1,status='old',file=wavelet_fn,access='sequential',&
       action='read',form='formatted',iostat=in_error)
  do jt = 1,nt
    read(unit=1,fmt=*,iostat=in_error) src(jt)
    if(in_error /= 0) exit
    nts_data = nts_data+1
  end do
  close(1)
else
  call sep_read(src,wavelet_fn)
!  do
!    if(src(nts_data+1) <= 0.) exit
!    nts_data = nts_data+1
!  end do
end if
call sourc_interp(dt,ntData,tIntFact,nt,src,s)
nts = tIntFact*nts_data
s(nts+1:nt) = 0
write(0,*) 's ',s(1:nts)
fNyq = 1/(2.0*dtData)
call from_param('fc',fc,fNyq)
!fc = min(fc,0.5*Vmin*SQRT(1.0/(dx*dx)+1.0/(dz*dz)))
fc = max(fc,0.5*Vmin/max(dx,dz))
jtStep = floor(1/(2*dt*fc))
if(jtStep < 1) jtStep = 1 
ntstep = floor(1.0*nt/jtStep)
!**************************************************************
!*** Initiate parameters for the REM or Lax-Wendroff method ***
!**************************************************************
call param_init(dt,dx,dz,nt,nz)
allocate(am(0:2),                   &
         STAT=allocation_status)
if(allocation_status /= 0) then
  write(0,*) 'rtm2d: Error! could not allocate REM arrays '
  call skull()
  stop 666
end if
call rtm_init(dt,am)
!*******************************************
!*** Obtain geometry of the shot gathers ***
!*******************************************
if(shot_blend) then
  nsources = nsuper_gath
  ngx = ngx_enc
else
  k = 1
  l = 1
  ngx_p = 1
  do jtraces = 1,ntraces
    if(jtraces < ntraces) then
      if(sx(jtraces) == sx(jtraces+1)) then
        k = k+1
      else
        if(k > ngx_p) then
          ngx = k
          ngx_p = k
        end if
        k = 1
        l = l+1
      end if
      nsources = l
    elseif(sx(1) > sx(ntraces)) then
      if(k < ngx) then
        k = k+1
      else
        if(k > ngx_p) then
          ngx = k
          ngx_p = k
        end if
        k = 1
        l = l+1
      end if
    end if
  end do
end if
if(nsources == 1) ngx = k 
napLDta = floor(apertureL/dx)
napRDta = floor(apertureR/dx)
napLMdl = floor(apertureL/dx)
napRMdl = floor(apertureR/dx)
nbox = ngx+napLMdl+napRMdl
write(0,*) 'ngx ',ngx

allocate(source_index(nsources,2),  &
         sgx(1-napLDta:ngx+napRDta+src_pad,nsources), &
         sgx_max(nsources), &
         rcv(ntData),&
         shot_gath(nt,ngx),  &
         STAT=allocation_status)
if(allocation_status /= 0) then
  write(0,*) 'Error! could not allocate sgx, shot_gath!'
  call skull()
  stop 666
end if

write(0,*) 'Prepare geometric arrays for either normal or blended sources'
sgx = UNDEFINED 
if(shot_blend) then
  k = 1
  l = 1
  do jtraces = 1,ntraces
    if(jtraces < ntraces) then
      if(super_gath(jtraces) == super_gath(jtraces+1)) then
        sgx(k,l) = gx(jtraces)
        k = k+1
      else
        source_index(l,1) = k
        source_index(l,2) = jtraces
        sgx(k,l) = gx(jtraces)
        k = 1
        l = l+1
      end if
    else
      sgx(k,l) = gx(jtraces)
      source_index(l,1) = k-1
      source_index(l,2) = jtraces

      if(k < ngx) then
        k = k+1
      else
        k = 1
        l = l+1
      end if
    end if
  end do  
else
  k = 1
  l = 1
  ngx_p = 1
  do jtraces = 1,ntraces
    if(jtraces < ntraces) then
      if(sx(jtraces) == sx(jtraces+1)) then
        sgx(k,l) = gx(jtraces)
        k = k+1
      else
        if(k > ngx_p) then
          ngx = k
          ngx_p = k
        end if
        source_index(l,1) = k
        source_index(l,2) = jtraces
        sgx(k,l) = gx(jtraces)
        k = 1
        l = l+1
      end if
    else
      sgx(k,l) = gx(jtraces)
      source_index(l,1) = k-1
      source_index(l,2) = jtraces

      if(k < ngx) then
        k = k+1
      else
        if(k > ngx_p) then
          ngx = k
          ngx_p = k
        end if
        k = 1
        l = l+1
      end if
    end if
  end do
end if
sgx_max = maxval(sgx(:,:),DIM=1)
do jgx = 1,src_pad+napRDta
  sgx(ngx+jgx,:) = sgx_max(:)+jgx*dx
end do
do jgx = 1-napLDta,0
  sgx(jgx,:) = sgx(1,:)-(1-jgx)*dx
end do
write(0,*) 'End geometric arrays for either normal or blended sources'
!*************************************************************
!* Prepare some Fourier quantities and initialize FFTW plans *
!*************************************************************
nxpad = nbox+pad+pad
nkgx = nxpad+src_pad
nkz = nz+pad+pad+zpad
!***************************
!* Verbose some parameters *
!***************************
write(0,*) '***************************'
write(0,*) '* Verbose some parameters *'
write(0,*) '***************************'
write(0,*) 'nz ',nz
write(0,*) 'start_jz ',start_jz
write(0,*) 'end_jz ',end_jz
write(0,*) 'ntData ',ntData
write(0,*) 'nt ',nt
write(0,*) 'ntstep ',ntstep
write(0,*) 'nts_data ',nts_data
write(0,*) 'nts ',nts
write(0,*) 'nx ',nx
write(0,*) 'ngx ',ngx
write(0,*) 'nkz ',nkz
write(0,*) 'nkgx ',nkgx
write(0,*) 'nbox ',nbox
write(0,*) 'ntraces ',ntraces
write(0,*) 'nsources ',nsources
write(0,*) 'Points triang filt ',npts
write(0,*) 'shot_blend ',shot_blend
write(0,*) 'first_cdp ',first_cdp
write(0,*) 'last_cdp ',last_cdp
write(0,*) 'src_pad',src_pad
write(0,*) 'Global av. vel. ',Vcte
write(0,*) 'Vel. min. ',Vmin
write(0,*) 'Vel. max. ',Vmax
write(0,*) 'Epsilon min. ',eps_min
write(0,*) 'Epsilon max. ',eps_max
write(0,*) 'Delta min. ',delta_min
write(0,*) 'Delta max. ',delta_max
write(0,*) 'Theta min. ',theta_min
write(0,*) 'Theta max. ',theta_max
write(0,*) 'Nyquist freq. ',fNyq
write(0,*) 'Mult. time step ',jtStep
write(0,*) 'Max. freq. in data ',fc
write(0,*) 'Vel. fracc. ',vel_fracc
write(0,*) 'dgx ',dgx
write(0,*) 'dx ',dx
write(0,*) 'dtData ',dtData
write(0,*) 'dt ',dt
write(0,*) 'dz ',dz
write(0,*) 'dgxOdx ',dgxOdx
write(0,*) 'Courant number ',cfl
write(0,*) 'tIntFact ',tIntFact
write(0,*) 'Aperture ',aperture
write(0,*) 'Aperture left',apertureL
write(0,*) 'Aperture right',apertureR
write(0,*) 'Use anis ',anis
write(0,*) 'Use tti ',tti
write(0,*) 'scalco ',scalco
write(0,*) 'nthreads ',nthreads
write(0,*) 'Snap fn',snap_fn
write(0,*) 'Snapr fn',snapr_fn
write(0,*) 'snapsht ',snapsht
!write(0,*) 'Energy imaging cond? ',energy
write(0,*) 'Use ispc? ',ispc
write(0,*) ' '
!***************************
!* Initialize FFTW plans ***
!***************************
!if(.not. energy) call ft1d_init(nkz)
!*********************************
!*** Define Fourier quantities ***
!*********************************
allocate(kz(nkz),         &
         fkzsep(nkz),     &
         kgx(nkgx),       &
         STAT=allocation_status)
if(allocation_status /= 0) then
  write(0,*) 'Error! could not allocate wavenumber vectors!'
  call skull()
  stop 666
end if

kgx_Nyq = pi/dx
kz_Nyq = pi/dz
dkgx = 2*kgx_Nyq/nkgx
dkz = 2*kz_Nyq/nkz
kz0 = -dkz
kzp = dkz
kgx = 0
kz = 0
k=-nkgx/2
do jgx = 1,nkgx
  kgx(jgx) = k*dkgx
  k = k+1
end do
k=-nkz/2
do jz = 1,nkz
  kz(jz) = k*dkz
  k = k+1
end do
kNyq = 2.0*pi*fNyq !/Vmin
!kNyq = SQRT(kz_Nyq**2+kgx_Nyq**2)
!kNyq = kz_Nyq
kcl = 0.0*kNyq
kcu = 1.0*kNyq

fkzsep = 1
do jz = 1,nkz
  if(kz(jz) < kz0) then
    fkzsep(jz) = 0
  else if(kz(jz) >= kz0 .and. kz(jz) <= kzp) then
    fkzsep(jz) = SIN(0.5*pi*(kz(jz)-kz0)/(kzp-kz0))**2
  end if
end do
!**************************************************
!*** Initialize REM solver and crosscorrelation ***
!**************************************************
if(shot_blend) then
  call weqsolver_init(2,nkz,nkgx,ngx,pad,anis,shot_blend,am,energy,&
                      nsx_enc,ngx_enc,ntraces,nphase,encod_vector)
else
  call weqsolver_init(2,nkz,nkgx,ngx,pad,anis,shot_blend,am,energy)
end if
call grad_init(ispc)
call xcorr_init(zpad+pad,fkzsep)
!***************************************************
!*** Initialize arrays for the main calculations ***
!***************************************************
allocate(pft    (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),&
         pftm1  (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
         pftm2  (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad), &
         prt    (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),    &
         prtp1  (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),&
         prtp2  (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),&
         qft    (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),    &
         qrt    (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),    &
         pS     (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+pad),    &
         num    (1:nz,             1-pad-napLMdl:ngx+napRMdl+pad),        &
         num0   (1:nz,             1-pad-napLMdl:ngx+napRMdl+pad),        &
         num90  (1:nz,             1-pad-napLMdl:ngx+napRMdl+pad),        &
         den    (1:nz,             1-pad-napLMdl:ngx+napRMdl+pad),        &
         den0   (1:nz,             1-pad-napLMdl:ngx+napRMdl+pad),        &
         denreal(1:nz,             1-pad-napLMdl:ngx+napRMdl+pad),        &
         psnap  (1-pad-zpad:nz+pad,first_cdp:last_cdp),&
         V_shot (1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad), &
         V_shotr(1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad), &
         alp    (1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad),&
       STAT=allocation_status)
if(allocation_status /= 0) then
  write(0,*) 'Error! could not allocate RTM/REM arrays!'
  call skull()
  stop 666
end if
allocate(wft   (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRDta+src_pad+pad),&
         wft1  (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRDta+src_pad+pad),&
         wft2  (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRDta+src_pad+pad),&
         wrt   (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRDta+src_pad+pad),&
         wrt1  (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRDta+src_pad+pad),&
         wrt2  (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRDta+src_pad+pad),&
         Vshot0(1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad))
!if(energy) then
  if(anis) then
         allocate(pftPX(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  pftRX(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  pftPZ(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  pftRZ(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  prtPX(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  prtRX(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  prtPZ(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  prtRZ(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  dpftP(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  dpftPm1(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  dpftPm2(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  dpftR(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  dpftRm1(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  dpftRm2(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  dprtP(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  dprtPp1(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  dprtPp2(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  dprtR(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  dprtRp1(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  dprtRp2(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  a(1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad),&
                  b(1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad),&
                  dd2(1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad),&
                  temp1(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  temp2(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad))
  else
         allocate(pftX(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  pftZ(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  prtX(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  prtZ(1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  dpft (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad),  &
                  dprt (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRMdl+src_pad+pad))
  end if
!end if
if(anis) then
  allocate(pftR  (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRDta+src_pad+pad),&
           wft1R (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRDta+src_pad+pad),&
           wft2R (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRDta+src_pad+pad),&
           prtR  (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRDta+src_pad+pad),&
           wrt1R (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRDta+src_pad+pad),&
           wrt2R (1-pad-zpad:nz+pad,1-pad-napLMdl:ngx+napRDta+src_pad+pad),&
           fanis1(1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad),&
           fanis2(1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad),&
           fanis3(1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad),&
           fanis4(1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad),&
           fanis5(1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad),&
           epsShot(1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad),&
           delShot(1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad),&
           lambP(1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad),&
           lambM(1-pad-zpad:nz+pad,1-pad:nbox+pad+src_pad),&
           STAT=allocation_status)
  if(allocation_status /= 0) then
    write(0,*) 'Error! could not allocate anisotrpy-related arrays!'
    call skull()
    stop 666
  end if
end if
!*****************************************************
!* Write files to the history file for the RTM image *
!*****************************************************
write(0,*) 'Writing history'
!call to_history("title",'RTM shot CIC',"CIC")
!call to_history("n1",nz,"CIC")
!call to_history("n2",nx,"CIC")
!call to_history("d1",dz,"CIC")
!call to_history("d2",dx,"CIC")
!call to_history("dz",dz,"CIC")
!call to_history("dcdp",dx,"CIC")
!call to_history("dline",1,"CIC")
!call to_history("o1",o1,"CIC")
!call to_history("o2",o2,"CIC")
!call to_history("label1",'depth',"CIC")
!call to_history("label2",'rec no.',"CIC")
!call to_history("unit1",unit1,"CIC")
!call to_history("unit2",unit2,"CIC")
!call to_history("esize",4,"CIC")

call to_history("title",'RTM shot CIC')
call to_history("n1",nz)
call to_history("n2",nx)
call to_history("d1",dz)
call to_history("d2",dx)
call to_history("dz",dz)
call to_history("dcdp",dx)
call to_history("dline",1)
call to_history("o1",o1)
call to_history("o2",o2)
call to_history("label1",'depth')
call to_history("label2",'rec no.')
call to_history("unit1",unit1)
call to_history("unit2",unit2)
call to_history("esize",4)

call to_history("title",'RTM shot EIC tomog.',"EIC1")
call to_history("n1",nz,"EIC1")
call to_history("n2",nx,"EIC1")
call to_history("d1",dz,"EIC1")
call to_history("d2",dx,"EIC1")
call to_history("dz",dz,"EIC1")
call to_history("dcdp",dx,"EIC1")
call to_history("dline",1,"EIC1")
call to_history("o1",o1,"EIC1")
call to_history("o2",o2,"EIC1")
call to_history("label1",'depth',"EIC1")
call to_history("label2",'rec no.',"EIC1")
call to_history("unit1",unit1,"EIC1")
call to_history("unit2",unit2,"EIC1")
call to_history("esize",4,"EIC1")

call to_history("title",'RTM shot EIC refl.',"EIC2")
call to_history("n1",nz,"EIC2")
call to_history("n2",nx,"EIC2")
call to_history("d1",dz,"EIC2")
call to_history("d2",dx,"EIC2")
call to_history("dz",dz,"EIC2")
call to_history("dcdp",dx,"EIC2")
call to_history("dline",1,"EIC2")
call to_history("o1",o1,"EIC2")
call to_history("o2",o2,"EIC2")
call to_history("label1",'depth',"EIC2")
call to_history("label2",'rec no.',"EIC2")
call to_history("unit1",unit1,"EIC2")
call to_history("unit2",unit2,"EIC2")
call to_history("esize",4,"EIC2")

call to_history('n1',nkz,"vrandom")
call to_history('n2',nkgx,"vrandom")
call to_history('n3',nsources,"vrandom")
call to_history('o1',(1-pad-zpad)*dz,"vrandom")
call to_history('o2',(1-pad)*dx,"vrandom")
call to_history('o3',1,"vrandom")
call to_history('d1',dz,"vrandom")
call to_history('d2',dx,"vrandom")
call to_history('d3',1,"vrandom")
if(anis) then
  call to_history('n1',nkz,"erandom")
  call to_history('n2',nkgx,"erandom")
  call to_history('n3',nsources,"erandom")
  call to_history('o1',(1-pad-zpad)*dz,"erandom")
  call to_history('o2',(1-pad)*dx,"erandom")
  call to_history('o3',1,"erandom")
  call to_history('d1',dz,"erandom")
  call to_history('d2',dx,"erandom")
  call to_history('d3',1,"erandom")

  call to_history('n1',nkz,"drandom")
  call to_history('n2',nkgx,"drandom")
  call to_history('n3',nsources,"drandom")
  call to_history('o1',(1-pad-zpad)*dz,"drandom")
  call to_history('o2',(1-pad)*dx,"drandom")
  call to_history('o3',1,"drandom")
  call to_history('d1',dz,"drandom")
  call to_history('d2',dx,"drandom")
  call to_history('d3',1,"drandom")
  if(th_f) then
    call to_history('n1',nkz,"trandom")
    call to_history('n2',nkgx,"trandom")
    call to_history('n3',nsources,"trandom")
    call to_history('o1',(1-pad-zpad)*dz,"trandom")
    call to_history('o2',(1-pad)*dx,"trandom")
    call to_history('o3',1,"trandom")
    call to_history('d1',dz,"trandom")
    call to_history('d2',dx,"trandom")
    call to_history('d3',1,"trandom")
  end if
end if
!*****************************************************
!*** Write the header files of every migrated shot ***
!*****************************************************
!write(0,*) 'Writing headers'
!call write_headers(cdp_no_key,x_key,  &
!       n_keys_snap,sx_key_snap,sy_key_snap,gx_key_snap,gy_key_snap,  &
!       hoff_key_snap,mute_key_snap,muts_key_snap,cdp_no_key_snap,line_no_key_snap,  &
!       nt,ntraces,nz,nx,x,sx,gx,first_cdp,last_cdp,&
!       z,wf_only)
!* I use this to generate an empty hff file
!call auxout("out")
!if(sep_put_number_keys("in",n_keys_vel) /= 0) &
!  write(0,*) 'Error, cannot write number of keys on the hff (.H@@) output file!'
!************************************************
!*** Initialize variables for shot velocities ***
!************************************************
call vshot_init(dx,dz,nz,nbox,first_cdp,last_cdp,&
           pad,zpad,src_pad,0.0,Vmax) 
!***********************************************
!*** Initialize variables for the main loops ***
!***********************************************
first_source = 1
last_source = nsources
jtraces = 1
if(one_src) then
  first_source = src_snap
  last_source = src_snap
  jtraces = 1+(src_snap-1)*ngx
end if
amp_rms = 0
amp_mean = 0
first_geof = 1
!**************************************************
!*** This is the main loop to migrate the shots ***
!**************************************************
I_corr = 0
I_tomo = 0
I_refl = 0
I0     = 0
I90    = 0
loop_over_sources: do jsources = first_source,last_source
  ngeof = source_index(jsources,1)
  last_geof = first_geof+ngeof-1
!**********************************************************************************
!*** Here I create velocity frames to migrate the shots, with random boundaries *** 
!**********************************************************************************
  call vshot(sgx(:,jsources),V_smth,V_shot,.true.,.false.) 
  call vshot(sgx(:,jsources),V_smth,V_shotr,.false.,.false.) 
  alp = 0
  call sep_write(V_shot,'vrandom')
  if(anis) then
!***************************************************************************************
!*** Anisotropy frames use no random boundaries. I have to test if they are required *** 
!***************************************************************************************
    fanis1=1
    fanis2=1
    fanis3=0
    fanis4=1
    epsShot=0
    delShot=0
    if(eps_f) then
      call vshot(sgx(:,jsources),Vanis1,epsShot,.false.,.true.)
      if(del_f) then
        call vshot(sgx(:,jsources),Vanis2,delShot,.false.,.true.) 
        where(epsShot < delShot)
          delShot = 0.99*epsShot
        end where
        fanis2 = SQRT(1.+2.*delShot)
        call sep_write(delShot,'drandom')
      end if
      fanis1 = 1+2.*epsShot
      call sep_write(epsShot,'erandom')
      if(th_f) then
        call vshot(sgx(:,jsources),theta, fanis3,.false.,.false.)
        fanis3 = sin(fanis3*deg2rad)
!        dummy=base_tris2d_oper(fanis3,20,20,10) 
        call vshot(sgx(:,jsources),theta, fanis4,.false.,.false.)
        fanis4 = cos(fanis4*deg2rad)
!        dummy=base_tris2d_oper(fanis4,20,20,10) 

        call sep_write(fanis3,'trandom')
!        call vshot(sgx(:,jsources),sinth, fanis3,.false.,.true.)
!        call vshot(sgx(:,jsources),costh, fanis4,.false.)
!        call vshot(sgx(:,jsources),sin2th,fanis5,.false.)
      end if
    end if
!    if(energy) then
          a = epsShot+SQRT((epsShot+1)**2-2.0*(epsShot-delShot))
          b = fanis2
          dd2 = a*a+b*b
!          lambP = 1.+(epsShot+tiny*eps_max)+SQRT((1.+epsShot+tiny*eps_max)**2-&
!                  2.*(epsShot+tiny*eps_max-delShot-tiny*delta_max))
!          lambM = 1.+(epsShot+tiny*eps_max)-SQRT((1.+epsShot+tiny*eps_max)**2-&
!                  2.*(epsShot+tiny*eps_max-delShot-tiny*delta_max))
          lambP = 1.+epsShot+SQRT((1.+epsShot)**2-&
                  2.*(epsShot-delShot))
          lambM = 1.+epsShot-SQRT((1.+epsShot)**2-&
                  2.*(epsShot+tiny*eps_max-delShot-tiny*delta_max))
!    end if

    if(tti) then
      call lap_init(V_shot,V_shotr,ispc,fanis3,fanis4) !,fanis5)
    else
      call lap_init(V_shot,V_shotr,ispc)
    end if
    write(0,*) 'Running with the finite differences method (anisotropic)!'
  else
    call lap_init(V_shot,V_shotr,ispc)
    write(0,*) 'Running with the finite differences  method!'
  end if

  shot_gath = 0

  if(shot_blend) then
    super_shots_pos(1) = 1+floor((sx(jtraces)-gx(jtraces))/dgx)
    do jencod = 2,nsx_enc
      jaux = (1+floor(1.0*ngx_enc/nsx_enc))
      dsx = sx(jtraces+jaux)-sx(jtraces)
      super_shots_pos(jencod) = super_shots_pos(jencod-1)+floor(dsx/dgx)
    end do
     write(0,*) ' '
     write(0,'(a,i7,i7,i7,i7,i7,i5)')     &
    'No. sources,curr source,source_index,first_geof,last_geof,ngeof',  &
     nsources,jsources,source_index(jsources,2),first_geof,last_geof,ngeof
  else
    shot_pos = 1+floor((sx(jtraces)-gx(jtraces))/dx)
     write(0,'(a,i7,i7,i7,i7,i7,i7,i5)')     &
    'No. sources,curr source,shot pos.,source_index,first_geof,last_geof,ngeof',  &
     nsources,jsources,shot_pos,source_index(jsources,2),first_geof,last_geof,ngeof
    write(0,*) 'sx,gx,jtraces',sx(jtraces),gx(jtraces),jtraces
  end if
!**************************
!*** Reading input data ***
!**************************
  if(dummy == 0) then
!**************************************
!*** Reading data (one shot gather) ***
!**************************************
    k = 1
    jgeof = ceiling(first_geof/dgxOdx)-1 
    do jaux = first_geof,last_geof,dgxOdx
      jgeof = jgeof+1
      previous_trace = jgeof-1
      if(sseek_block("datafile",previous_trace,4*ntData,0) < 0) &
      call seperr('rtm2d: Have a trouble seeking trace')
      if(sreed("datafile",rcv,4*ntData) /= 4*ntData) &
      call seperr('rtm2d: Have a trouble reading trace')
      call rcv_interp(ntData,tIntFact,nt,rcv,shot_gath(:,k),.false.)
      k = k+1
    end do
  end if
!*************************************************************************
!*** Lax-Wendroff or pseudo-analytical method for the source wavefield ***
!*************************************************************************
!  if(energy) then
    pftm2 = 0
    pftm1 = 0
    prtp2 = 0
    prtp1 = 0
!  end if
  pft = 0
  pft(1,shot_pos) = s(1)
!  pft(2,shot_pos) = -s(1)
  prt = 0
  jtCoarse = 0 
  qft = 0
  qft = alp*pft
  wft2 = 0
  wft1 = pft-alp*pft 
  wft = pft-qft 
  if(anis) then
    dpftPm1 = 0 
    dpftPm2 = 0 
    dpftP = wft
    dpftRm1 = 0 
    dpftRm2 = 0 
    dpftR = wft
    dprtPp1 = 0 
    dprtPp2 = 0 
    dprtP   = 0 
    dprtRp1 = 0 
    dprtRp2 = 0 
    dprtR   = 0 
  else
    dpft = wft
  end if
  if(dgxOdx == 1) then
    prt(1,1:ngeof) = shot_gath(nt,1:ngeof)
    prt(2,1:ngeof) = -shot_gath(nt,1:ngeof)
  else
    do jgx = 1,ngx
      jfx = 1+(jgx-1)/dgxOdx
      jx = jfx
      jfx = jfx-jx
      if(jx >= 1 .and. jx < ngeof) & 
        prt(1,jgx) = (1.0-jfx)*shot_gath(nt,jx)+jfx*shot_gath(nt,jx+1)
        prt(2,jgx) = -(1.0-jfx)*shot_gath(nt,jx)-jfx*shot_gath(nt,jx+1)
    end do
  end if
  if(anis) then
    prtR = 0
    if(dgxOdx == 1) then
      prtR(1,1:ngeof) = shot_gath(nt,1:ngeof)
      prtR(2,1:ngeof) = -shot_gath(nt,1:ngeof)
    else
      do jgx = 1,ngx
        jfx = 1+(jgx-1)/dgxOdx
        jx = jfx
        jfx = jfx-jx
        if(jx >= 1 .and. jx < ngeof) & 
          prtR(1,jgx) = (1.0-jfx)*shot_gath(nt,jx)+jfx*shot_gath(nt,jx+1)
          prtR(2,jgx) = -(1.0-jfx)*shot_gath(nt,jx)-jfx*shot_gath(nt,jx+1)
      end do
    end if
  end if
  num   = 0
  num0  = 0
  num90 = 0
  den   = 0
  den0  = 0
! **********************************************************
! *** Source propagation with random boundary conditions ***
! **********************************************************
  if(anis) then
    write(0,*) 'Source wavefield forward (anisotropy mode):'
    wft2R = 0
    wft1R = 0
    pftR = 0 
    pftR(1,shot_pos) = s(1) 
!    pftR(2,shot_pos) = -s(1) 
    loop_over_time_An: do jt = 2,nt
      call fwd_prop_anis(wft1,wft2,pft,wft1R,wft2R,pftR,ispc,fanis1,fanis2,&
                         dpftPm1,dpftPm2,dpftP,dpftRm1,dpftRm2,dpftR,tti,energy)
      pft(1,shot_pos)  = pft(1,shot_pos) +s(jt)
!      pft(2,shot_pos)  = pft(2,shot_pos) -s(jt)
      pftR(1,shot_pos) = pftR(1,shot_pos)+s(jt)
!      pftR(2,shot_pos) = pftR(2,shot_pos)-s(jt)
      if(mod(jt,jtStep) == 0) then
        jtCoarse = jtCoarse+1
      end if
    end do loop_over_time_An
    write(0,*) 'Source and receiver wavefields backward (anisotropy mode):'
    wrt2 = 0
    wrt1 = 0
    wrt2R = 0
    wrt1R = 0
    wft2=wft1
    wft1= pft
    pft = wft2
    wft2R=wft1R
    wft1R= pftR
    pftR = wft2R
    dpftPm2 = dpftPm1
    dpftPm1 = dpftP
    dpftP   = dpftPm2
    dpftRm2 = dpftRm1
    dpftRm1 = dpftR
    dpftR   = dpftRm2
!***********************************************************************
!*** Source and receiver propagation with random boundary conditions ***
!***********************************************************************
    loop_over_time2_An: do jt = nt-1,1,-1
!      wft(1,shot_pos) = wft(1,shot_pos)+s(jt)
      call bwd_propRB_anis(wrt1,wrt2,prt,wrt1R,wrt2R,prtR,ispc,fanis1,fanis2,&
                      dprtPp1,dprtPp2,dprtP,dprtRp1,dprtRp2,dprtR,&
                      wft1,wft2,pft,wft1R,wft2R,pftR,&
                      dpftPm1,dpftPm2,dpftP,dpftRm1,dpftRm2,dpftR,tti,energy)
!***********************************************************
!*** Damping function only applied to the reverse field, ***
!***********************************************************
      call taperings(wrt2,wrt1,prt)
      call taperings(wrt2R,wrt1R,prtR)
!      if(energy)
      call taperings(dprtPp2,dprtPp1,dprtP)
!      if(energy)
      call taperings(dprtRp2,dprtRp1,dprtR)
      if(dgxOdx == 1) then
        prt(1,1:ngeof)  = shot_gath(jt,1:ngeof)
!        prt(2,1:ngeof)  = -shot_gath(jt,1:ngeof)
        prtR(1,1:ngeof) = shot_gath(jt,1:ngeof)
!        prtR(2,1:ngeof) = -shot_gath(jt,1:ngeof)
      else
        do jgx = 1,ngx
          jfx = 1+(jgx-1)/dgxOdx
          jx = jfx
          jfx = jfx-jx
          if(jx >= 1 .and. jx < ngeof) & 
            prt(1,jgx)    = (1.0-jfx)*shot_gath(jt,jx)+jfx*shot_gath(jt,jx+1)
            prt(2,jgx)    = -(1.0-jfx)*shot_gath(jt,jx)-jfx*shot_gath(jt,jx+1)
            prtR(1,jgx)   = (1.0-jfx)*shot_gath(jt,jx)+jfx*shot_gath(jt,jx+1)
            prtR(2,jgx)   = -(1.0-jfx)*shot_gath(jt,jx)-jfx*shot_gath(jt,jx+1)
        end do
      end if
      pft(1,shot_pos) = pft(1,shot_pos)+s(jt)
!      pft(2,shot_pos) = pft(2,shot_pos)-s(jt)
      pftR(1,shot_pos) = pftR(1,shot_pos)+s(jt)
!      pftR(2,shot_pos) = pftR(2,shot_pos)-s(jt)

      if(mod(jt,jtStep) == 0) then
        jtCoarse = jtCoarse-1
        if (snapsht .and. jsources == src_snap .AND. mod(jtCoarse,jtsnap) == 0) then
          ntSnap = ntSnap+1
          loop_srcwave_rb2An:do jgx = 1-napLDta,ngx+napRDta,1
            jx = nint(1.5001+sgx(jgx,jsources)/dx)
            if(jx >= first_cdp .and. jx <= last_cdp) then
              psnap(:,jx) = pft(:,jgx)
            end if
          end do loop_srcwave_rb2An 
          loop_srcwave_rb3An:do jx = first_cdp,last_cdp 
            call sep_write(psnap(1:nz,jx),snap_fn)
          end do loop_srcwave_rb3An

          loop_srcwave_rb2aAn:do jgx = 1-napLDta,ngx+napRDta,1
            jx = nint(1.5001+sgx(jgx,jsources)/dx)
            if(jx >= first_cdp .and. jx <= last_cdp) then
              psnap(:,jx) = prt(:,jgx)
            end if
          end do loop_srcwave_rb2aAn
          loop_srcwave_rb3aAn:do jx = first_cdp,last_cdp 
            call sep_write(psnap(1:nz,jx),snapr_fn)
          end do loop_srcwave_rb3aAn
        end if
! *** Obtain the components of the gradient of forward wavefield 
!        if(energy) then
          do jgx = 1-napLMdl-pad,ngx+napRMdl+pad+src_pad,1
            temp1(:,jgx) = fanis4(:,jgx+napLMdl)*pft(:,jgx)
          end do
          do jgx = 1-napLMdl-pad,ngx+napRMdl+pad+src_pad,1
            temp2(:,jgx) = fanis3(:,jgx+napLMdl)*pft(:,jgx)
          end do
          call grad(temp1,temp2,ispc) 
          pftPX=temp1-temp2

          do jgx = 1-napLMdl-pad,ngx+napRMdl+pad+src_pad,1
            temp1(:,jgx) = fanis3(:,jgx+napLMdl)*pftR(:,jgx)
          end do
          do jgx = 1-napLMdl-pad,ngx+napRMdl+pad+src_pad,1
            temp2(:,jgx) = fanis4(:,jgx+napLMdl)*pftR(:,jgx)
          end do
          call grad(temp1,temp2,ispc) 
          pftRZ=temp1+temp2

          do jgx = 1-napLMdl-pad,ngx+napRMdl+pad+src_pad,1
            temp1(:,jgx) = fanis4(:,jgx+napLMdl)*prt(:,jgx)
          end do
          do jgx = 1-napLMdl-pad,ngx+napRMdl+pad+src_pad,1
            temp2(:,jgx) = fanis3(:,jgx+napLMdl)*prt(:,jgx)
          end do
          call grad(temp1,temp2,ispc) 
          prtPX=temp1-temp2

          do jgx = 1-napLMdl-pad,ngx+napRMdl+pad+src_pad,1
            temp1(:,jgx) = fanis3(:,jgx+napLMdl)*prtR(:,jgx)
          end do
          do jgx = 1-napLMdl-pad,ngx+napRMdl+pad+src_pad,1
            temp2(:,jgx) = fanis4(:,jgx+napLMdl)*prtR(:,jgx)
          end do
          call grad(temp1,temp2,ispc) 
          prtRZ=temp1+temp2
!        end if
!**********************************************************
!*** Imaging condition after random boundary conditions ***
!**********************************************************
!        if(energy) then
!$omp parallel do default(shared) &
!$omp private(jgx,jx) 
          do jgx = 1-napLMdl,ngx+napRMdl,1
            jx = jgx+napLMdl
            num0(:,jgx) =  num0(:,jgx)+pftPX(1:nz,jgx)*prtPX(1:nz,jgx)+pftRZ(1:nz,jgx)*prtRZ(1:nz,jgx)+&
                                    ((a(1:nz,jx)*dpftP(1:nz,jgx)+b(1:nz,jx)*dpftR(1:nz,jgx))*&
                                     (a(1:nz,jx)*dprtP(1:nz,jgx)+b(1:nz,jx)*dprtR(1:nz,jgx))/&
                                     lambP(1:nz,jx)+&
                                     (-b(1:nz,jx)*dpftP(1:nz,jgx)+a(1:nz,jx)*dpftR(1:nz,jgx))*&
                                     (-b(1:nz,jx)*dprtP(1:nz,jgx)+a(1:nz,jx)*dprtR(1:nz,jgx))/&
                                     lambM(1:nz,jx))/(V_shotr(1:nz,jx)**2*dd2(1:nz,jx))
            num90(:,jgx) = num90(:,jgx)+pftPX(1:nz,jgx)*prtPX(1:nz,jgx)+pftRZ(1:nz,jgx)*prtRZ(1:nz,jgx)+&
                            cos2thc*((a(1:nz,jx)*dpftP(1:nz,jgx)+b(1:nz,jx)*dpftR(1:nz,jgx))*&
                                     (a(1:nz,jx)*dprtP(1:nz,jgx)+b(1:nz,jx)*dprtR(1:nz,jgx))/&
                                     lambP(1:nz,jx)+&
                                     (-b(1:nz,jx)*dpftP(1:nz,jgx)+a(1:nz,jx)*dpftR(1:nz,jgx))*&
                                     (-b(1:nz,jx)*dprtP(1:nz,jgx)+a(1:nz,jx)*dprtR(1:nz,jgx))/&
                                     lambM(1:nz,jx))/(V_shotr(1:nz,jx)**2*dd2(1:nz,jx))
            den0(:,jgx) = den0(:,jgx)+pftPX(1:nz,jgx)*pftPX(1:nz,jgx)+pftRZ(1:nz,jgx)*pftRZ(1:nz,jgx)+&
                                    ((a(1:nz,jx)*dpftP(1:nz,jgx)+b(1:nz,jx)*dpftR(1:nz,jgx))*&
                                     (a(1:nz,jx)*dpftP(1:nz,jgx)+b(1:nz,jx)*dpftR(1:nz,jgx))/&
                                     lambP(1:nz,jx)+&
                                     (-b(1:nz,jx)*dpftP(1:nz,jgx)+a(1:nz,jx)*dpftR(1:nz,jgx))*&
                                     (-b(1:nz,jx)*dpftP(1:nz,jgx)+a(1:nz,jx)*dpftR(1:nz,jgx))/&
                                     lambM(1:nz,jx))/(V_shotr(1:nz,jx)**2*dd2(1:nz,jx))
          end do
!$omp end parallel do
!        else
!!$omp parallel do default(shared) &
!!$omp private(jgx) 
          do jgx = 1-napLMdl,ngx+napRMdl,1
            call xcorr(pft(:,jgx),prt(:,jgx),num(:,jgx),den(:,jgx))
          end do
!!$omp end parallel do
!        end if
      end if
    end do loop_over_time2_An
      dtSnap = dt*jtsnap
  else
    write(0,*) 'Source wavefield forward:'
    loop_over_time: do jt = 2,nt
      call fwd_prop(wft1,wft2,wft,qft,ispc,pftm1,pftm2,dpft,energy)
      pft = wft/(1.0-alp)
      pft(1,shot_pos) = pft(1,shot_pos)+s(jt)
!      pft(2,shot_pos) = pft(2,shot_pos)-s(jt)
      qft = alp*pft
      wft = pft-qft 
      if(mod(jt,jtStep) == 0) then
        jtCoarse = jtCoarse+1
      end if
    end do loop_over_time
    write(0,*) 'Source and receiver wavefields backward:'
    qrt = alp*prt
    wrt2 = 0
    wrt1 = 0
    wrt = prt-qrt 
    wft2=wft1
    wft1= wft
    wft = wft2
    pft=wft/(1.0-alp)
    qft=alp*pft
!    if(energy) then
      pftm2=pftm1
      pftm1= dpft
      dpft = pftm2
!    end if
! ***********************************************************************
! *** Source and receiver propagation with random boundary conditions ***
! ***********************************************************************
    loop_over_time2: do jt = nt-1,1,-1
!        wft(1,shot_pos) = wft(1,shot_pos)+s(jt)
      call bwd_propRB(wrt1,wrt2,wrt,qrt,ispc,&
                      prtp1,prtp2,dprt,wft1,wft2,wft,qft,pftm1,pftm2,dpft,energy)
!***********************************************************
!*** Damping function only applied to the reverse field, ***
!***********************************************************
      call taperings(wrt2,wrt1,wrt)
!      if(energy) 
      call taperings(prtp2,prtp1,dprt)
      prt = wrt/(1.0-alp)
      if(dgxOdx == 1) then
        prt(1,1:ngeof) = shot_gath(jt,1:ngeof)
!        prt(2,1:ngeof) = -shot_gath(jt,1:ngeof)
      else
        do jgx = 1,ngx
          jfx = 1+(jgx-1)/dgxOdx
          jx = jfx
          jfx = jfx-jx
          if(jx >= 1 .and. jx < ngeof) & 
            prt(1,jgx)   = (1.0-jfx)*shot_gath(jt,jx)+jfx*shot_gath(jt,jx+1)
            prt(2,jgx)   = -(1.0-jfx)*shot_gath(jt,jx)-jfx*shot_gath(jt,jx+1)
        end do
      end if
      qrt = alp*prt
      wrt = prt-qrt 
      pft = wft/(1.0-alp)
      pft(1,shot_pos) = pft(1,shot_pos)+s(jt)
!      pft(2,shot_pos) = pft(2,shot_pos)-s(jt)
      qft = alp*pft
      wft = pft-qft

      if(mod(jt,jtStep) == 0) then
        jtCoarse = jtCoarse-1
        if (snapsht .and. jsources == src_snap .AND. mod(jtCoarse,jtsnap) == 0) then
          ntSnap = ntSnap+1
          loop_srcwave_rb2:do jgx = 1-napLDta,ngx+napRDta,1
            jx = nint(1.5001+sgx(jgx,jsources)/dx)
            if(jx >= first_cdp .and. jx <= last_cdp) then
              psnap(:,jx) = pft(:,jgx)
            end if
          end do loop_srcwave_rb2 
          loop_srcwave_rb3:do jx = first_cdp,last_cdp 
            call sep_write(psnap(1:nz,jx),snap_fn)
          end do loop_srcwave_rb3

          loop_srcwave_rb2a:do jgx = 1-napLDta,ngx+napRDta,1
            jx = nint(1.5001+sgx(jgx,jsources)/dx)
            if(jx >= first_cdp .and. jx <= last_cdp) then
              psnap(:,jx) = prt(:,jgx)
            end if
          end do loop_srcwave_rb2a
          loop_srcwave_3a:do jx = first_cdp,last_cdp 
            call sep_write(psnap(1:nz,jx),snapr_fn)
          end do loop_srcwave_3a
        end if
! *** Obtain the components of the gradient of forward wavefield 
!        if(energy) then
          pftX=pft 
          pftZ=pft
          call grad(pftX,pftZ,ispc) 
          prtX=prt
          prtZ=prt
          call grad(prtX,prtZ,ispc) 
!        end if
!**********************************************************
!*** Imaging condition after random boundary conditions ***
!**********************************************************
        loop_crosscorrelation:do jgx = 1-napLMdl,ngx+napRMdl,1
!          if(energy) then
            num0(:,jgx)  = num0(:,jgx)+pftX(1:nz,jgx)*prtX(1:nz,jgx)+&
                                    pftZ(1:nz,jgx)*prtZ(1:nz,jgx)+&
                                    dpft(1:nz,jgx)*dprt(1:nz,jgx)/&
                                    V_shotr(1:nz,jgx+napLMdl)**2
            num90(:,jgx) = num90(:,jgx)+pftX(1:nz,jgx)*prtX(1:nz,jgx)+&
                                    pftZ(1:nz,jgx)*prtZ(1:nz,jgx)+&
                            cos2thc*dpft(1:nz,jgx)*dprt(1:nz,jgx)/&
                                    V_shotr(1:nz,jgx+napLMdl)**2
            den0(:,jgx) = den0(:,jgx)+pftX(1:nz,jgx)*pftX(1:nz,jgx)+&
                                    pftZ(1:nz,jgx)*pftZ(1:nz,jgx)+&
                                    dpft(1:nz,jgx)*dpft(1:nz,jgx)/&
                                    V_shotr(1:nz,jgx+napLMdl)**2
!          else
            call xcorr(pft(:,jgx),prt(:,jgx),num(:,jgx),den(:,jgx))
!          end if
        end do loop_crosscorrelation
      end if
    end do loop_over_time2
    dtSnap = dt*jtsnap
  end if
!*****************************************
!*** Imaging condition: Time summation ***
!*****************************************
!*** 2D Triangle filter according to Guitton et al. (2006)
    denreal=real(den)
    dummy=base_tris2d_oper(denreal,npts,npts,10) 
    den=denreal
    denmax=maxval(abs(den))
    denreal=real(den0)
    dummy=base_tris2d_oper(denreal,npts,npts,10) 
    den0=denreal
    denmax0=maxval(abs(den0))
    do jgx = 1-napLDta,ngx+napRDta,1
      jfx = nint(1.5001+(sgx(jgx,jsources)-0.0)/dx) !-sgx(1-napLDta,jsources))/dx
      jx = jfx
      jfx = jfx-jx 
      if(jx >= first_cdp .and. jx < last_cdp) then
!        if(energy) then
!          if(cos2thc == 1) then
            I0(:,jx)    = I0 (:,jx)+(1-jfx)*num0(:, jgx)/(den0(:,jgx)+tiny*denmax0)
            I0(:,jx+1)  = I0 (:,jx+1)+ jfx *num0(:, jgx)/(den0(:,jgx)+tiny*denmax0)
            I90(:,jx)   = I90(:,jx)+(1-jfx)*num90(:,jgx)/(den0(:,jgx)+tiny*denmax0)
            I90(:,jx+1) = I90(:,jx+1)+ jfx *num90(:,jgx)/(den0(:,jgx)+tiny*denmax0)
!          else
!            I90(:,jx)   = I90(:,jx)+(1-jfx)*num90(:,jgx)/(den0(:,jgx)+tiny*denmax)
!            I90(:,jx+1) = I90(:,jx+1)+ jfx *num90(:,jgx)/(den0(:,jgx)+tiny*denmax)
!          end if
!        else
          I_corr(:,jx)   = I_corr(:,jx)+(1-jfx)*num(:,jgx)/(den(:,jgx)+tiny*denmax)
          I_corr(:,jx+1) = I_corr(:,jx+1)+ jfx *num(:,jgx)/(den(:,jgx)+tiny*denmax)
!        end if
        do jz = start_jz,end_jz
          if(isnan(I_corr(jz,jx)) .OR. isnan(I0(jz,jx)) .OR. isnan(I90(jz,jx))) then
            write(0,*) 'Error! I_corr, I0 or I90 is a NaN in shot! ',jsources,'jz,jx = ',jz,jx
            write(0,*) 'num, numb0, numb90, den, den0: ',&
                        num(jz,jgx),num0(jz,jgx),num90(jz,jgx),den(jz,jgx),den0(jz,jgx)
            call skull()
            stop 666
          end if
        end do
      end if
    end do
!    if(energy) then
      temp1=exp(-alpha*abs(I90)/(abs(I0)+abs(I90)+tiny*maxval(abs(I0)+abs(I90))))
!      if(cos2thc == 1) then
        I_tomo = I0 !*temp1
!      else
        I_refl = I90 !*(1-temp1)
!      end if
!    end if
!  end if
  jtraces_ind(jsources) = ngeof
  jtraces = jtraces+ngeof
  first_geof = last_geof+1
  call clean_all(tti,ispc)
end do loop_over_sources
!************************************
!*** Finish RTM, write the output ***
!************************************
if(.not. (one_src .and. snapsht)) then
  min_amplitude = 99999.0  
  max_amplitude = -99999.0
  n_nonzero_samples = 0
  do jx = first_cdp,last_cdp
    do jz = start_jz,end_jz
      if(isnan(I_corr(jz,jx))) then
        write(0,*) 'Error! I_corr is a NaN! jz,jx = ',jz,jx
        call skull()
        stop 666
      end if
  
      if(min_amplitude > I_corr(jz,jx)) min_amplitude = I_corr(jz,jx)
      if(max_amplitude < I_corr(jz,jx)) max_amplitude = I_corr(jz,jx)
      if(I_corr(jz,jx) /= 0) then
        amp_rms = amp_rms+I_corr(jz,jx)*I_corr(jz,jx)
        amp_mean = amp_mean+I_corr(jz,jx)
        n_nonzero_samples = n_nonzero_samples+1
      end if
    end do
    call sep_write(I_corr(:,jx))
  end do
  
  if(n_nonzero_samples == 0) then
    write(0,*) 'I_corr is made of zeros',n_nonzero_samples
    call skull()
    stop 666
  end if
  amp_rms = SQRT(amp_rms/n_nonzero_samples)
  amp_mean = amp_mean/n_nonzero_samples
  call to_history("rms_amplitude",amp_rms)
  call to_history("mean_amplitude",amp_mean)
  call to_history("min_amplitude",min_amplitude)
  call to_history("max_amplitude",max_amplitude)
  call to_history("nonzero_samples",n_nonzero_samples)
  write(0,*) '******CIC Stack section:*******'
  write(0,*) 'Non-zero samples',n_nonzero_samples
  write(0,*) 'rms_amplitude',amp_rms
  write(0,*) 'mean_amplitude',amp_mean
  write(0,*) 'max_amplitude',max_amplitude
  write(0,*) 'min_amplitude',min_amplitude
  write(0,*) '**************************'

  min_amplitude = 99999.0  
  max_amplitude = -99999.0
  n_nonzero_samples = 0
  do jx = first_cdp,last_cdp
    do jz = start_jz,end_jz
      if(isnan(I_tomo(jz,jx))) then
        write(0,*) 'Error! I_tomo is a NaN! jz,jx = ',jz,jx
        call skull()
        stop 666
      end if
  
      if(min_amplitude > I_tomo(jz,jx)) min_amplitude = I_tomo(jz,jx)
      if(max_amplitude < I_tomo(jz,jx)) max_amplitude = I_tomo(jz,jx)
      if(I_tomo(jz,jx) /= 0) then
        amp_rms = amp_rms+I_tomo(jz,jx)*I_tomo(jz,jx)
        amp_mean = amp_mean+I_tomo(jz,jx)
        n_nonzero_samples = n_nonzero_samples+1
      end if
    end do
    call sep_write(I_tomo(:,jx),"EIC1")
  end do
  
  if(n_nonzero_samples == 0) then
    write(0,*) 'I_tomo is made of zeros',n_nonzero_samples
    call skull()
    stop 666
  end if
  amp_rms = SQRT(amp_rms/n_nonzero_samples)
  amp_mean = amp_mean/n_nonzero_samples
  call to_history("rms_amplitude",amp_rms)
  call to_history("mean_amplitude",amp_mean)
  call to_history("min_amplitude",min_amplitude)
  call to_history("max_amplitude",max_amplitude)
  call to_history("nonzero_samples",n_nonzero_samples)
  write(0,*) '****** EIC1 (tomog. comp.) Stack section:*******'
  write(0,*) 'Non-zero samples',n_nonzero_samples
  write(0,*) 'rms_amplitude',amp_rms
  write(0,*) 'mean_amplitude',amp_mean
  write(0,*) 'max_amplitude',max_amplitude
  write(0,*) 'min_amplitude',min_amplitude
  write(0,*) '**************************'

  n_nonzero_samples = 0
  do jx = first_cdp,last_cdp
    do jz = start_jz,end_jz
      if(isnan(I_refl(jz,jx))) then
        write(0,*) 'Error! I_refl is a NaN! jz,jx = ',jz,jx
        call skull()
        stop 666
      end if
  
      if(min_amplitude > I_refl(jz,jx)) min_amplitude = I_refl(jz,jx)
      if(max_amplitude < I_refl(jz,jx)) max_amplitude = I_refl(jz,jx)
      if(I_refl(jz,jx) /= 0) then
        amp_rms = amp_rms+I_refl(jz,jx)*I_refl(jz,jx)
        amp_mean = amp_mean+I_refl(jz,jx)
        n_nonzero_samples = n_nonzero_samples+1
      end if
    end do
    call sep_write(I_refl(:,jx),"EIC2")
  end do
  
  if(n_nonzero_samples == 0) then
    write(0,*) 'I_refl is made of zeros',n_nonzero_samples
    call skull()
    stop 666
  end if
  amp_rms = SQRT(amp_rms/n_nonzero_samples)
  amp_mean = amp_mean/n_nonzero_samples
  call to_history("rms_amplitude",amp_rms)
  call to_history("mean_amplitude",amp_mean)
  call to_history("min_amplitude",min_amplitude)
  call to_history("max_amplitude",max_amplitude)
  call to_history("nonzero_samples",n_nonzero_samples)
  write(0,*) '****** EIC2 (refl. comp.) Stack section:*******'
  write(0,*) 'Non-zero samples',n_nonzero_samples
  write(0,*) 'rms_amplitude',amp_rms
  write(0,*) 'mean_amplitude',amp_mean
  write(0,*) 'max_amplitude',max_amplitude
  write(0,*) 'min_amplitude',min_amplitude
  write(0,*) '**************************'
end if

if(snapsht) then
  call to_history('o1',(start_jz-1)*dz,snap_fn)
  call to_history('o2',(first_cdp-1)*dx,snap_fn)
  call to_history('o3',0,snap_fn)
  call to_history('d1',dz,snap_fn)
  call to_history('d2',dx,snap_fn)
  call to_history('d3',dtSnap,snap_fn)
  call to_history('n1',nz,snap_fn)
  call to_history('n2',nx,snap_fn)
  call to_history('n3',ntSnap,snap_fn)
  call to_history('o1',(start_jz-1)*dz,snapr_fn)
  call to_history('o2',(first_cdp-1)*dx,snapr_fn)
  call to_history('o3',0,snapr_fn)
  call to_history('d1',dz,snapr_fn)
  call to_history('d2',dx,snapr_fn)
  call to_history('d3',dtSnap,snapr_fn)
  call to_history('n1',nz,snapr_fn)
  call to_history('n2',nx,snapr_fn)
  call to_history('n3',ntSnap,snapr_fn)
end if

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------
write(0,*) ' '
write(0,*) 'Finishing the rtm2d program normally!'
!call sep_close()
end program rtm2d
